@@ -490,7 +490,6 @@ void cFemonOsd::DrawInfoWindow(void)
                          OSDDRAWINFOLEFT(tr("Protocol"),   *data.protocol);
                          offset += OSDROWHEIGHT;
                          OSDDRAWINFOLEFT(tr("Bitrate"),    *data.bitrate);
-                         offset += OSDROWHEIGHT;
                          }
                       }
                    }
@@ -626,6 +625,23 @@ void cFemonOsd::Action(void)
               m_BERValid = (ioctl(m_Frontend, FE_READ_BER, &m_BER) >= 0);
               m_UNCValid = (ioctl(m_Frontend, FE_READ_UNCORRECTED_BLOCKS, &m_UNC) >= 0);
               }
+           else if (strstr(*cDevice::ActualDevice()->DeviceType(), SATIP_DEVICE)) {
+              m_Quality = cDevice::ActualDevice()->SignalQuality();
+              m_QualityValid = (m_Quality >= 0);
+              m_Strength = cDevice::ActualDevice()->SignalStrength();
+              m_StrengthValid = (m_Strength >= 0);
+              m_FrontendName = cDevice::ActualDevice()->DeviceName();
+              m_FrontendStatus = (fe_status_t)(cDevice::ActualDevice()->HasLock() ? (FE_HAS_LOCK | FE_HAS_SIGNAL | FE_HAS_CARRIER | FE_HAS_VITERBI | FE_HAS_SYNC) : 0);
+              m_FrontendStatusValid = m_StrengthValid;
+              m_Signal = uint16_t(m_Strength * 0xFFFF / 100);
+              m_SignalValid = m_StrengthValid;
+              m_SNR = uint16_t(m_Quality * 0xFFFF / 100);
+              m_SNRValid = m_QualityValid;
+              m_BER = 0;
+              m_BERValid = false;
+              m_UNC = 0;
+              m_UNCValid = false;
+              }
            else if (m_SvdrpConnection.handle >= 0) {
               cmd.handle = m_SvdrpConnection.handle;
               m_SvdrpPlugin->Service("SvdrpCommand-v1.0", &cmd);
@@ -716,6 +732,9 @@ void cFemonOsd::Show(void)
            return;
            }
         }
+     else if (strstr(*cDevice::ActualDevice()->DeviceType(), SATIP_DEVICE)) {
+        // nop
+        }
      else if (femonConfig.usesvdrp) {
         if (!SvdrpConnect() || !SvdrpTune())
            return;
@@ -790,6 +809,9 @@ void cFemonOsd::ChannelSwitch(const cDevice * device, int channelNumber, bool li
            return;
            }
         }
+     else if (strstr(*cDevice::ActualDevice()->DeviceType(), SATIP_DEVICE)) {
+        // nop
+        }
      else if (femonConfig.usesvdrp) {
         if (!SvdrpConnect() || !SvdrpTune())
            return;
@@ -836,26 +858,83 @@ bool cFemonOsd::DeviceSwitch(int direction)
      cChannel *channel = Channels.GetByNumber(cDevice::CurrentChannel());
      if (channel) {
         for (int i = 0; i < cDevice::NumDevices() - 1; i++) {
-           if (direction >= 0) {
-              if (++device >= cDevice::NumDevices())
-                 device = 0;
-              }
-           else {
-              if (--device < 0)
-                 device = cDevice::NumDevices() - 1;
-              }
-           if (cDevice::GetDevice(device)->ProvidesChannel(channel, 0)) {
-              debug("%s(%d) device(%d)\n", __PRETTY_FUNCTION__, direction, device);
-              cStatus::MsgChannelSwitch(cDevice::PrimaryDevice(), 0, true);
-              cControl::Shutdown();
-              cDevice::GetDevice(device)->SwitchChannel(channel, true);
-              if (cDevice::GetDevice(device) == cDevice::PrimaryDevice())
-                 cDevice::GetDevice(device)->ForceTransferMode();
-              cControl::Launch(new cTransferControl(cDevice::GetDevice(device), channel));
-              cStatus::MsgChannelSwitch(cDevice::PrimaryDevice(), channel->Number(), true);
-              return (true);
-              }
-           }
+            if (direction >= 0) {
+               if (++device >= cDevice::NumDevices())
+                  device = 0;
+               }
+            else {
+               if (--device < 0)
+                  device = cDevice::NumDevices() - 1;
+               }
+            // Collect the current priorities of all CAM slots that can decrypt the channel:
+            int NumCamSlots = CamSlots.Count();
+            int SlotPriority[NumCamSlots];
+            int NumUsableSlots = 0;
+            bool NeedsDetachAllReceivers = false;
+            bool InternalCamNeeded = false;
+            bool ValidDevice = false;
+            cCamSlot *s = NULL;
+            cDevice *d = cDevice::GetDevice(device);
+            if (channel->Ca() >= CA_ENCRYPTED_MIN) {
+               for (cCamSlot *CamSlot = CamSlots.First(); CamSlot; CamSlot = CamSlots.Next(CamSlot)) {
+                   SlotPriority[CamSlot->Index()] = MAXPRIORITY + 1; // assumes it can't be used
+                   if (CamSlot->ModuleStatus() == msReady) {
+                      if (CamSlot->ProvidesCa(channel->Caids())) {
+                         if (!ChannelCamRelations.CamChecked(channel->GetChannelID(), CamSlot->SlotNumber())) {
+                            SlotPriority[CamSlot->Index()] = CamSlot->Priority();
+                            NumUsableSlots++;
+                            }
+                         }
+                      }
+                  }
+               if (!NumUsableSlots)
+                  InternalCamNeeded = true; // no CAM is able to decrypt this channel
+               }
+            for (int j = 0; j < NumCamSlots || !NumUsableSlots; ++j) {
+                if (NumUsableSlots && SlotPriority[j] > MAXPRIORITY)
+                   continue; // there is no CAM available in this slot
+                bool HasInternalCam = d->HasInternalCam();
+                if (InternalCamNeeded && !HasInternalCam)
+                   continue; // no CAM is able to decrypt this channel and the device uses vdr handled CAMs
+                if (NumUsableSlots && !HasInternalCam && !CamSlots.Get(j)->Assign(d, true))
+                   continue; // CAM slot can't be used with this device
+                if (d->ProvidesChannel(channel, 0, &NeedsDetachAllReceivers)) { // this device is basically able to do the job
+                   debug("%s(%d) device(%d)\n", __PRETTY_FUNCTION__, direction, device);
+                   if (NumUsableSlots && !HasInternalCam && d->CamSlot() && d->CamSlot() != CamSlots.Get(j))
+                      NeedsDetachAllReceivers = true; // using a different CAM slot requires detaching receivers
+                   if (NumUsableSlots && !HasInternalCam)
+                      s = CamSlots.Get(j);
+                   ValidDevice = true;
+                   break;
+                   }
+                if (!NumUsableSlots)
+                   break; // no CAM necessary, so just one loop over the devices
+                }
+            // Do the actual switch if valid device found
+            if (d && ValidDevice) {
+               cStatus::MsgChannelSwitch(cDevice::PrimaryDevice(), 0, true);
+               cControl::Shutdown();
+               if (NeedsDetachAllReceivers)
+                  d->DetachAllReceivers();
+               if (s) {
+                  if (s->Device() != d) {
+                     if (s->Device())
+                        s->Device()->DetachAllReceivers();
+                     if (d->CamSlot())
+                        d->CamSlot()->Assign(NULL);
+                     s->Assign(d);
+                     }
+                  }
+               else if (d->CamSlot() && !d->CamSlot()->IsDecrypting())
+                  d->CamSlot()->Assign(NULL);
+               d->SwitchChannel(channel, true);
+               if (d == cDevice::PrimaryDevice())
+                  d->ForceTransferMode();
+               cControl::Launch(new cTransferControl(d, channel));
+               cStatus::MsgChannelSwitch(cDevice::PrimaryDevice(), channel->Number(), true);
+               return (true);
+               }
+            }
         }
      }
    return (false);
